/*
  Java Japanese-English dictionary/Kanjidic browser
  Copyright (C) 1997 Todd David Rudick

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */

import java.util.*;
import java.awt.*;
import java.io.*;
import java.awt.event.*;
import GUI;
import ByteSearch;
import SkipGUI;
import GenList;
import KanjiViewer;
//import HWRGUI;
import HighLightMap;
import CopyLeft;
import BushuGUI;
//import OnlineDocs;
import Filter;
import ClearState;

public class JavaDict {
  //  static public Frame f;
  static String JDICPREFIX = "JDic: ";
  static String MENUSTRING = "$Revision: 1.1 $$Date: 1997/07/20 07:24:58 $"+" Main Menu";
  static String EDICTSTRING = "Japanese -> English Dictionary";
  static String KDICTSKIP = "Kanjidict / Skip-code lookup";
  static String KDICTBUSH = "Kanjidict / Bushu lookup";
  static String KANJIVIEWSTRING = "Kanji Viewer";
  static String KDICTHWR = "Kanjidict / Handwriting Recognition";
  static String OPTIONS = "Options";
  static String COPYLEFT = "Copying: terms of distribution";
  static String ONLINE = "Online Documentation";
  Frame quitframe=null;
  public Font DFont;
  public String DFontName;
  public int DFontStyle;
  public int DFontSize;
  public Font BigFont;
  public Font BiggerFont;
  public KanjiViewer kanjiview;
  public Frame frame;
  public boolean useRomaji;

  private final static String MenuDirections = "Menu...";

  static public class FlipPage implements ActionListener {
    Frame f;
    String page;
    JavaDict javadict;

    public FlipPage(JavaDict javadict, Frame f, String page) {
      this.f = f;
      this.page = page;
      this.javadict = javadict;
    }
    public void doaction() {
      if (javadict.quitframe!=null) {
	try {
	  javadict.quitframe.dispose();
	} catch (Exception e) { }
      }
      CardLayout clayout = (CardLayout) f.getLayout();
      clayout.show(f,page);
      f.setTitle(JDICPREFIX + page);
    }
    public void actionPerformed(ActionEvent e) {
      doaction();
    }
  }
  
  static class LookUp implements Runnable {
    static Boolean lock = new Boolean(false);
    String pattern;
    GenList output;
    String dict, dictindex;
    static final int MAXLINE = 320;
    static final int BUFSIZE = 4096*2;
    boolean messages;
    public static ThreadGroup lookupgroup = new ThreadGroup("LookUp");
    Vector postFilters;

    
    public LookUp(String pattern, GenList output, 
		  String dict, String dictindex, boolean messages,
		  JavaDict javadict) {
      this.postFilters=new Vector();
      if (javadict.optiongui.getRomajiFlag()) 
	this.postFilters.addElement
	  (new Filter.romajiFilter(GUI.Rom2KanaVector));
      this.pattern=pattern;
      this.output=output;
      this.dict = dict;
      this.dictindex = dictindex;
      this.messages = messages;
    }
    /*    public LookUp(String pattern, GenList output,
	  String dict, String dictindex, boolean messages,
	  Vector postFilters) {
	  this.postFilters=postFilters;
	  this.pattern=pattern;
	  this.output=output;
	  this.dict = dict;
	  this.dictindex = dictindex;
	  this.messages = messages;
	  }*/

    public void run() {
      //      System.out.println("Entering monitor");
      synchronized (lock) {
	//	System.out.println("Got monitor");
	try { // catch all exceptions to avoid monitor lock race condition
	  output.removeAll();
	  if (messages) {
	    if (dictindex!=null) 
	      output.add("Searching file \"" + dict + "\", using index \"" + 
			 dictindex + "\" for: \"" + pattern+"\"");
	    else 
	      output.add("Searching file \"" + dict + 
			 "\" for: \"" + pattern+"\"");
	  }
	
	  // Convert pattern into EUC
	  //	  CharToByteConverter fromUnicode = GUI.fromUnicode;
	  //	  ByteToCharConverter toUnicode = GUI.toUnicode;
	  
	  byte[] EUCpattern;
	  try {
	    EUCpattern=pattern.getBytes("EUCJIS");
	      //fromUnicode.convertAll(pattern.toCharArray());
	  } catch (Exception e) {
	    if (messages)
	      output.add("Error: Could not convert pattern"+
			 " to EUC encoding.\n");
	    /*	    else 
		    System.out.println("Error: Could not convert "+
		    "pattern to EUC"+
		    " encoding.\n");*/
	    return;
	  }
	  if (EUCpattern.length==0) {
	    if (messages)
	      output.add("Error: Could not convert pattern"+
			 " to EUC encoding.\n");
	    /*	    else 
		    System.out.println("Error: Could not convert"+
		    " pattern to EUC"+
		    " encoding.\n");*/
	    return;
	  }
	  if (dictindex==null) { 	// perform grep through file
	    FileInputStream f;
	    // open the file
	    try {
	      f=new FileInputStream(dict);
	    } catch (Exception e) {
	      if (messages)
		output.add("Error: can't open dictionary.");
	      else System.out.println("Error: can't open dictionary.");
	      return;
	    }
	    
	    // allocate buffers
	    byte[][] Buffer = new byte[2][BUFSIZE+EUCpattern.length+10];
	    int BufLen=0;
	    int curbuf = 0;
	    // so that it seems there is a newline at
	    // the start of the file ... ->
	    Buffer[1][BUFSIZE-1-(EUCpattern.length-1)]=012; 
	    
	    boolean finished = false;
	    try {
	      BufLen=f.read(Buffer[curbuf],0,BUFSIZE);
	      if (BufLen!=BUFSIZE) finished=true;
	    } catch (IOException e) { System.out.println(e); }
	    
	    // allocate a BM - search object
	    ByteSearch bmsearch = new ByteSearch(EUCpattern);
	    int index=0;
	    Thread mythread = java.lang.Thread.currentThread();
	    do {
	      while (index!=-1) {
		if (mythread.activeCount()>1) {
		  //System.out.println("Search Interrupted. Leaving monitor");
		  return;
		}
		index=bmsearch.search(Buffer[curbuf],index,BUFSIZE);
		if (index!=-1) {
		  // Got a hit
		  // find the index of the preceding newline + 1
		  int hitpos = 0; /* <- to keep track of where in the final 
				     string
				     it ends up */
		  int previndex = index;
		  byte[] prevbuf = Buffer[curbuf];
		  while (prevbuf[previndex]!=012) {
		    previndex--;
		    hitpos++;
		    if (previndex<0) {
		      prevbuf = Buffer[1-curbuf];
		      previndex = BUFSIZE-1-(EUCpattern.length-1);
		    }
		  } 
		  // + 1
		  previndex++;
		  hitpos--;
		  if ((previndex>BUFSIZE-1-(EUCpattern.length-1))&&
		      (prevbuf == Buffer[1-curbuf])) {
		    previndex=0;
		    prevbuf = Buffer[curbuf];
		  }
		  
		  // start building the output line from that point on
		  byte[] Line = new byte[MAXLINE*2];
		  int Linei=0;
		  while (prevbuf[previndex]!=012) {
		    Line[Linei]=prevbuf[previndex];
		    Linei++;
		    previndex++;
		    if ((previndex>BUFSIZE-1-(EUCpattern.length-1))&&
			(prevbuf == Buffer[1-curbuf])) { /* if we're still 
							    behind */
		      previndex=0;
		      prevbuf = Buffer[curbuf];
		    } else if (previndex>BufLen) { // if we need to read
		      try {
			if (finished) {
			  if (messages)
			    output.add("Search Finished");
			  //	 System.out.println("Leaving monitor");
			  return;
			}
			int rbytes;
			rbytes = f.read(Buffer[curbuf],0,BUFSIZE);
			if (rbytes!=BUFSIZE) finished=true;
			BufLen = rbytes;
		      } catch (IOException e) { System.out.println(e); }
		      previndex=0;
		    }
		  }
		  byte tb=0;
		  if (hitpos>0) tb = Line[hitpos-1];
		  if ((pattern.charAt(0)>'\u30FF') // starts with kanji
		      ||(hitpos==0) // start of line
		      ||(tb==040) // space
		      ||(tb==057) // slash
		      ||(tb==0133) // square bracket
		      ||(tb==0174) // curly bracket
		      ||(tb==055) // hyphen
		      ||(tb==050) // paren
		      ||(tb==042) // double quote
		      ||(tb==047) // single quote
		      ||(tb==0140) // single grave quote 
		      ||(tb==0174)) // vert. bar
		    {
		      // convert the line back to UNICODE and add to output
		      String SLine;
		      try {
			/*
			  char[] carray = new char[Linei];
			  int numout;
			  numout=
			  toUnicode.convert(Line,0,Linei-1,carray,
			  0,Linei-1);
			  SLine = new String(carray,0,numout);*/
			SLine = new String(Line,0,Linei,"EUCJIS");
			SLine = SLine.trim();

		      } catch (Exception e5) {
			//System.out.println("Error converting output line.");
			continue;
		      }
		      // Check to see if the UNICODE pattern is really in the
		      // line.. It might not be if the match wasn't on an even
		      // EUC character
		      if (SLine.indexOf(pattern)!=-1)
 {
			Enumeration e = postFilters.elements();
			while ((e.hasMoreElements())&&(SLine.length()>0)) {
			  SLine=((Filter)(e.nextElement())).filter(SLine);
			}
			if (SLine.length()>0) 
			  output.add(SLine);
		      }
		    } //else output.add("Rejected hit hitpos = " + hitpos);
		  
		  index=previndex;
		}
	      } /* so index==-1 now, so shuffle stuff around */
	      if (finished) break;
	      int i;
	      int newbuf = 1-curbuf;
	      int mystart=BUFSIZE-EUCpattern.length+1;
	      for (i=mystart;i<BUFSIZE;i++)
		Buffer[newbuf][i-mystart]=Buffer[curbuf][i];
	      index=0;
	      curbuf=newbuf;
	      try {
		int rbytes;
		rbytes=f.read(Buffer[curbuf],BUFSIZE-mystart,mystart);
		BufLen=rbytes+(BUFSIZE-mystart);
		if (BufLen!=BUFSIZE) finished=true;
	      } catch (IOException e) { System.out.println(e);}
	    } while (true);
	  }
	  if (messages)
	    output.add("Search Finished");
	} catch (Exception e) { 
	  System.out.println("Search interrupted by exception:"+
			     e.getMessage());
	}
      } // synchronized(lock);
      //System.out.println("Out of monitor");
    } // run()
  }
  public GUI gui;
  public JavaDict.FlipPage todictionary;
  public JavaDict.FlipPage tokanjiviewer;
  public OptionGUI optiongui;

  static public void main(String args[]) {
    Button MenuButton;
    // TODO make a splash

    JavaDict javadict = new JavaDict();

    javadict.useRomaji=false;
    Frame f = new Frame(JDICPREFIX);
    f.setForeground(Color.black);
    f.setBackground(Color.lightGray);
    javadict.frame = f;
    f.setLayout(new CardLayout());

    javadict.DFont = f.getFont();
    if (javadict.DFont==null) {
      f.pack();
      javadict.DFont = f.getFont();
    }

    javadict.DFontName = "Sans-serif";

    javadict.DFontStyle = javadict.DFont.getStyle();


    javadict.DFontSize = javadict.DFont.getSize();
    javadict.BigFont = new Font(javadict.DFontName,
				javadict.DFontStyle,
				//javadict.DFontSize*4/3);
				17);
    javadict.BiggerFont = new Font(javadict.DFontName,
				   javadict.DFontStyle,
				   //javadict.DFontSize*2);
				   23);

    // order of the following is vital! some values are needed during the
    // initialization of what follows (e.g. GUI copies javadict.kanjiview)

    MenuButton = new Button(MenuDirections);
    MenuButton.addActionListener(new FlipPage(javadict, f,MENUSTRING));
    CopyLeft cleft = new CopyLeft(" This program, the handwriting recognition"+
				  " database, and all associated\n"+
				  "utilities and documentation are to be "+
				  "considered a part of this package,\n"+
				  "which is hereby released under the terms "+
				  "which follow.\n"+
				  "  The Japanese-English dictionary and "+
				  "Kanji dictionary fall under seperate\n"+
				  "copyrights. Consult the documentation "+
				  "that comes with those files for more\n"+
				  "information.\n",MenuButton);

    f.add(cleft,COPYLEFT); 

    /*  took out online docs
    MenuButton = new Button(MenuDirections);
    MenuButton.addActionListener(new FlipPage(javadict, f,MENUSTRING));
    OnlineDocs old = new OnlineDocs(MenuButton);
    f.add(old,ONLINE);
    */

    MenuButton = new Button(MenuDirections);
    MenuButton.addActionListener(new FlipPage(javadict, f,MENUSTRING));
    javadict.kanjiview = new KanjiViewer(javadict, MenuButton,
					 "kanjidic.sorted", null);
    javadict.tokanjiviewer = new FlipPage(javadict, 
					  f, KANJIVIEWSTRING);

    f.add(javadict.kanjiview,KANJIVIEWSTRING);

    MenuButton = new Button(MenuDirections);
    MenuButton.addActionListener(new FlipPage(javadict, f,MENUSTRING));
    javadict.gui = new GUI(javadict,MenuButton,
			   "edict", null /*"edict.xjdx"*/);
    f.add(javadict.gui,EDICTSTRING);

    MenuButton = new Button(MenuDirections);
    MenuButton.addActionListener(new FlipPage(javadict, f,MENUSTRING));
    SkipGUI skipgui = new SkipGUI(javadict, MenuButton,
				  "kanjidic.sorted",
				  null /* skip.idx? */);
    f.add(skipgui,KDICTSKIP);

    MenuButton = new Button(MenuDirections);
    MenuButton.addActionListener(new FlipPage(javadict, f,MENUSTRING));
    javadict.optiongui = new OptionGUI(javadict, MenuButton);
    f.add(javadict.optiongui,OPTIONS);


    MenuButton = new Button(MenuDirections);
    MenuButton.addActionListener(new FlipPage(javadict, f,MENUSTRING));
    HWRGUI hwrgui = new HWRGUI(javadict,MenuButton);
    f.add(hwrgui,KDICTHWR);

    MenuButton = new Button(MenuDirections);
    MenuButton.addActionListener(new FlipPage(javadict, f,MENUSTRING));
    BushuGUI bushugui = new BushuGUI(javadict,MenuButton,"kanjidic.sorted",null);
    f.add(bushugui,KDICTBUSH);

    javadict.optiongui.addClearState(skipgui);
    javadict.optiongui.addClearState(bushugui);

    MenuGUI menugui = new MenuGUI(javadict, f);
    f.add(menugui,MENUSTRING);
    
    f.pack();
    f.show();

    FlipPage fp = new FlipPage(javadict, f,MENUSTRING);
    fp.doaction();
  }

  static class MenuGUI extends Panel {

    
    static class quitok implements ActionListener {
      static class QuitOkL implements ActionListener {
	boolean yesbut;
	JavaDict javadict;
	public QuitOkL(JavaDict javadict, boolean yesbut) {
	  this.yesbut = yesbut;
	  this.javadict = javadict;
	}
	public void actionPerformed(ActionEvent e) {
	  if (yesbut) System.exit(0);
	  else {
	    try {
	      javadict.quitframe.dispose();
	    } catch (Exception e23) { }
	  }
	}
      }

      Frame f;
      JavaDict javadict;
      public quitok(JavaDict javadict, Frame f) {
	this.f=f;
	this.javadict=javadict;
      }
      public void actionPerformed(ActionEvent e) {
	Frame w = new Frame("Really Quit?");
	if (javadict.quitframe!=null) {
	  try {
	    javadict.quitframe.dispose();
	  } catch (Exception e23) { }
	}
	javadict.quitframe = w;
	w.setLayout(new GridLayout(2,1));
	Panel p1 = new Panel();
	w.add(p1);
	Panel p2 = new Panel();
	w.add(p2);
	Label l = new Label("Are you sure you want to exit?");
	p1.add(l);
	p2.setLayout(new FlowLayout(FlowLayout.CENTER,40,3));
	Button b1 = new Button("Yes");
	b1.addActionListener(new QuitOkL(javadict,true));
	p2.add(b1,"West");
	Button b2 = new Button("No");
	b2.addActionListener(new QuitOkL(javadict,false));
	p2.add(b2,"East");
	w.pack();
	w.show();
      }
    }



    public MenuGUI(JavaDict javadict, Frame f) {
      setBackground(Color.white);
      javadict.todictionary = new JavaDict.FlipPage(javadict, f, EDICTSTRING);
      this.setLayout(new BorderLayout());

      // try to use HighlightMap (if gifs are unavailable fall back on buttons)
      try {
        byte [] imageb1;
	byte [] imageb2;
	File f1 = new File("title1.gif");
	File f2 = new File("title2.gif");
	int f1len = (int)f1.length();
	int f2len = (int)f2.length();
	imageb1 = new byte[f1len];
	imageb2 = new byte[f2len];
	int index=0;
	FileInputStream fi1 = new FileInputStream(f1);
	FileInputStream fi2 = new FileInputStream(f2);
	while (index<f1len) index+=fi1.read(imageb1,index,f1len-index);
	index=0;
	while (index<f2len) index+=fi2.read(imageb2,index,f2len-index);
	Image i1 = Toolkit.getDefaultToolkit().createImage(imageb1);
        Image i2 = Toolkit.getDefaultToolkit().createImage(imageb2);

        HighLightMap m = new HighLightMap(i1,i2,600,200);
	m.addActionListener(new JavaDict.FlipPage(javadict,f,COPYLEFT),
			    new Rectangle(128,64,138,78),"Gnu GPL");
	m.addActionListener(javadict.todictionary,
			    new Rectangle(273,10,320,24),"JE");
	m.addActionListener(new JavaDict.FlipPage(javadict,f,KDICTSKIP),
			    new Rectangle(281,58,134,26),"Skip");
	m.addActionListener(new JavaDict.FlipPage(javadict,f,KDICTBUSH),
			    new Rectangle(281,86,156,21),"Bushu");
	m.addActionListener(new JavaDict.FlipPage(javadict,f,KDICTHWR),
			    new Rectangle(281,113,263,24),"HWR");
	m.addActionListener(new JavaDict.FlipPage(javadict,f,OPTIONS),
			    new Rectangle(275,144,85,24),"Options");

	m.addActionListener(new JavaDict.MenuGUI.quitok(javadict,f),
			    new Rectangle(274,171,52,25),"Quit");
                            
        Panel buttonPanel = new Panel();
        add(buttonPanel,"East");
        buttonPanel.add(m);

      } catch (IOException ioe384) {

	
	Panel titlePanel = new Panel(new GridLayout(2,1));
	this.add(titlePanel,"West");
	titlePanel.setFont(javadict.BigFont);
	titlePanel.add(new Label("Gnu",Label.CENTER));
	titlePanel.add(new Label("JavaDict",Label.CENTER));

	Panel buttonPanel = new Panel();
	add(buttonPanel,"East");
        buttonPanel.setLayout(new GridLayout(3,3));
	// create and add a button for each application
	Button edict = new Button("Japanese - English Dictionary");
	Button kanjskip = new Button("Kanji Lookup - Skip Method");
        Button kanjbush = new Button("Kanji Lookup - Bushu Method");
	Button kanjhwr = new Button("Kanji Lookup - Handwriting");
	Button options = new Button("Options");
	//	Button help = new Button("On-line Documentation");
	Button copying = new Button("Copyleft Notice");
	Button quit = new Button("Quit...");
        

	buttonPanel.add(edict);
	edict.addActionListener(javadict.todictionary);
	
	buttonPanel.add(kanjskip);
	kanjskip.addActionListener(new JavaDict.FlipPage(javadict,f,
							 KDICTSKIP));
        buttonPanel.add(kanjbush);
        kanjbush.addActionListener(new JavaDict.FlipPage(javadict,f,
                                                         KDICTBUSH));

	
	buttonPanel.add(kanjhwr);
	kanjhwr.addActionListener(new JavaDict.FlipPage(javadict,f,KDICTHWR));
	buttonPanel.add(options);
	options.addActionListener(new JavaDict.FlipPage(javadict,f,OPTIONS));

	
	//buttonPanel.add(help);
	//help.addActionListener(new JavaDict.FlipPage(javadict,f,ONLINE));
	
	buttonPanel.add(copying);
	copying.addActionListener(new JavaDict.FlipPage(javadict,f,
							COPYLEFT));

	buttonPanel.add(quit);
	quit.addActionListener(new JavaDict.MenuGUI.quitok(javadict,f));

      //      Button tempkanjiview = new Button("Kanjiviewer");
      //      this.add(tempkanjiview);
      //      tempkanjiview.addActionListener(javadict.tokanjiviewer);
      }
    }
  }
}



